/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.FileDescriptor;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.nio.channels.Channels;
import java.util.Collections;
import java.util.EnumSet;
import java.util.Objects;
import java.util.Set;

import org.imca_cat.prgm.internal.UnbufferedWritableByteChannelWriter;

/**
 * A skeletal computer program.
 * <p>
 * The only method that must be overridden in a subclass is
 * {@link Program#run() run()}; this is the entry point of the program.
 * <p>
 * This program synchronizes critical sections on the object returned by
 * {@link #lock()}; subclasses that need to synchronize their own critical
 * sections on the same object can obtain it via this method.
 * <p>
 * Here's an example "Hello World" program:
 *
 * <pre>
 * public class HelloWorld {
 *   public static void main(String[] args) {
 *     new AbstractProgram() {
 *       &#064;Override
 *       public int run() throws Exception {
 *         System.out.println("hello, world");
 *         return 0;
 *       }
 *     }.runAndExitVm();
 *   }
 * }</pre>
 * <p>
 * This class is thread-safe.
 */
public abstract class AbstractProgram implements Program {
  private final Object lock;
  private final PrintWriter err;
  private final Set<Option> options;

  /**
   * Constructs an instance with an {@link Object} critical section lock, an
   * error stream that is a {@link PrintWriter} on the system standard error
   * stream without buffering, without automatic flushing, and using the
   * default character encoding, and an empty options set.
   */
  protected AbstractProgram() {
    /* Unbuffered. */
    this(new Object(),
        new PrintWriter(new UnbufferedWritableByteChannelWriter(
            Channels.newChannel(new FileOutputStream(FileDescriptor.err)))),
        EnumSet.noneOf(Option.class));
  }

  /**
   * Constructs an instance with the specified critical section lock, an error
   * stream that is a {@link PrintWriter} on the system standard error stream
   * without buffering, without automatic flushing, and using the default
   * character encoding, and an empty options set.
   *
   * @param lock the object on which this program's critical sections
   *          synchronize
   *
   * @throws NullPointerException if {@code lock} is {@code null}
   */
  protected AbstractProgram(Object lock) {
    /* Unbuffered. */
    this(lock,
        new PrintWriter(new UnbufferedWritableByteChannelWriter(
            Channels.newChannel(new FileOutputStream(FileDescriptor.err)))),
        EnumSet.noneOf(Option.class));
  }

  /**
   * Constructs an instance with the specified critical section lock, error
   * stream, and options set.
   *
   * @param lock the object on which this program's critical sections
   *          synchronize
   * @param err the stream where error output should be written; this program
   *          assumes ownership of the stream; the client should not maintain
   *          references to it and must not use or modify it in any way; if the
   *          stream is thread-safe, then it is safe for the client to use it,
   *          but output to it by this program cannot be guaranteed to be
   *          correct unless the client maintains a reference to the
   *          {@code lock} object and synchronizes on it when using the stream
   * @param options the set of options controlling how this program behaves;
   *          this program assumes ownership of the set; the client should not
   *          maintain references to it and must not modify it in any way
   *
   * @throws NullPointerException if {@code lock}, {@code err}, or
   *           {@code options} is {@code null}
   */
  protected AbstractProgram(Object lock, PrintWriter err, Set<Option> options) {
    super();
    this.lock = Objects.requireNonNull(lock, "lock is null");
    this.err = Objects.requireNonNull(err, "err is null");
    this.options = Collections.unmodifiableSet(Objects.requireNonNull(options, "options is null"));
  }

  /**
   * {@inheritDoc}
   *
   * @implSpec This implementation calls {@link #lock()} and does the following
   *           while synchronized on the returned object: prints a stack trace
   *           of the specified exception to the error stream and flushes it.
   *           If the specified exception is a {@link ProgramDieException}, a
   *           stack trace is not printed, and instead the exception's message,
   *           if not {@code null} and not empty, is printed followed by the
   *           platform's line separator string. If this program was created
   *           with the {@link Option#LOCALIZED LOCALIZED} option, the
   *           {@linkplain ProgramDieException#getLocalizedMessage() localized
   *           message} of the {@code ProgramDieException} is used; otherwise
   *           the {@linkplain ProgramDieException#getMessage() non-localized
   *           message} is used.
   */
  @Override
  public void printError(Throwable t) {
    synchronized (lock()) {
      if (t instanceof ProgramDieException) {
        String m = options.contains(Option.LOCALIZED) ? t.getLocalizedMessage() : t.getMessage();
        if (m != null && !m.isEmpty()) err.println(m);
      } else {
        t.printStackTrace(err);
      }
      err.flush();
    }
  }

  /**
   * {@inheritDoc}
   *
   * @implSpec This implementation returns {@link Program#EXIT_FAILURE} unless
   *           the specified exception is a {@link ProgramDieException}, in
   *           which case {@linkplain ProgramDieException#exitStatus() its exit
   *           status} is returned.
   */
  @Override
  public int exitStatus(Throwable t) {
    if (t instanceof ProgramDieException) {
      return ((ProgramDieException)t).exitStatus();
    } else {
      return Program.EXIT_FAILURE;
    }
  }

  /**
   * {@inheritDoc}
   *
   * @implSpec This implementation calls {@link #lock()} and does the following
   *           while synchronized on the returned object: flushes the error
   *           stream by calling its {@link PrintWriter#flush() flush()}
   *           method.
   */
  @Override
  public void flush() {
    synchronized (lock()) {
      err.flush();
    }
  }

  /**
   * Runs this program and exits the VM with the exit status of this program.
   *
   * @implSpec This implementation invokes
   *           {@link Programs#runAndExitVm(Program)} with this program as the
   *           argument.
   */
  public void runAndExitVm() {
    Programs.runAndExitVm(this);
  }

  /**
   * Returns the object on which this program's critical sections synchronize.
   *
   * @implSpec This implementation returns the critical section lock object
   *           this program was constructed with.
   *
   * @return the critical section lock object
   */
  protected Object lock() {
    return lock;
  }

  /**
   * Options controlling the behavior of an {@link AbstractProgram}.
   */
  public static enum Option {
    /** Prefer localized messages. */
    LOCALIZED
  }
}
