/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.util.Locale;

/**
 * A {@link SensibleUi} that can be configured.
 * <p>
 * Configuration occurs by converting this UI into a builder on which mutations
 * can be performed and then converting the builder back into a new UI instance
 * that can replace the original.
 */
public interface ConfigurableSensibleUi extends SensibleUi {
  /**
   * Converts this UI into a builder.
   * <p>
   * The conversion between a UI and a builder is lossless. For example, if a
   * default value is used in a UI because it was not explicitly set in the
   * builder from which it was created, the builder returned by invoking this
   * method on that UI would again consider the value to not be set.
   *
   * @return a builder initialized from this UI
   */
  Builder toBuilder();

  /**
   * A <i>builder</i> for creating {@link ConfigurableSensibleUi} instances.
   */
  public static interface Builder {
    /**
     * Sets the name of the program.
     *
     * @param programName the name of the program; may be {@code null}; must
     *          not be an empty string
     *
     * @return this builder
     *
     * @throws IllegalArgumentException if {@code programName} is an empty
     *           string
     */
    Builder programName(String programName);

    /**
     * Sets the console used for prompting the user.
     *
     * @param console the console; may be {@code null}
     *
     * @return this builder
     */
    Builder console(ProgramConsole console);

    /**
     * Sets the standard output stream.
     *
     * @param stdout the standard output stream
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code stdout} is {@code null}
     */
    Builder stdout(Appendable stdout);

    /**
     * Sets the standard error stream.
     *
     * @param stderr the standard error stream
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code stderr} is {@code null}
     */
    Builder stderr(Appendable stderr);

    /**
     * Sets whether the standard output stream should be automatically flushed.
     * <p>
     * if {@code true}, methods of the built UI that write to the standard
     * output stream will flush the standard output stream after writing to it.
     *
     * @param autoFlush whether to automatically flush the standard output
     *          stream
     *
     * @return this builder
     */
    Builder autoFlushStdout(boolean autoFlush);

    /**
     * Sets whether the standard error stream should be automatically flushed.
     * <p>
     * if {@code true}, methods of the built UI that write to the standard
     * error stream will flush the standard error stream after writing to it.
     *
     * @param autoFlush whether to automatically flush the standard error
     *          stream
     *
     * @return this builder
     */
    Builder autoFlushStderr(boolean autoFlush);

    /**
     * Sets the locale to use for locale-sensitive operations.
     *
     * @param locale the locale
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code locale} is {@code null}
     */
    Builder locale(Locale locale);

    /**
     * Sets the {@code quiet} level flag.
     * <p>
     * If {@code true}, status messages will <i>not</i> be printed.
     *
     * @param quiet the {@code quiet} level flag
     *
     * @return this builder
     */
    Builder quiet(boolean quiet);

    /**
     * Sets the {@code verbose} level flag.
     * <p>
     * If {@code true}, note messages will be printed.
     *
     * @param verbose the {@code verbose} level flag
     *
     * @return this builder
     */
    Builder verbose(boolean verbose);

    /**
     * Sets the {@code debug} level flag.
     * <p>
     * If {@code true}, debug messages will be printed.
     *
     * @param debug the {@code debug} level flag
     *
     * @return this builder
     */
    Builder debug(boolean debug);

    /**
     * Sets whether exception stack traces should be printed.
     *
     * @param printStackTrace whether to print stack traces
     *
     * @return this builder
     */
    Builder printStackTrace(boolean printStackTrace);

    /**
     * Sets whether to use localized or non-localized exception messages when
     * printing messages from exceptions.
     *
     * @param useLocalizedExceptionMessage whether to use localized or
     *          non-localized exception messages
     *
     * @return this builder
     */
    Builder useLocalizedExceptionMessage(boolean useLocalizedExceptionMessage);

    /**
     * Selects which of this builder's output streams status messages should be
     * written to.
     *
     * @param statusTo which output stream status messages should be written
     *          to; must be {@code StandardIo.OUTPUT} or
     *          {@code StandardIo.ERROR}
     *
     * @return this builder
     *
     * @throws IllegalArgumentException if {@code statusTo} is not
     *           {@code StandardIo.OUTPUT} or {@code StandardIo.ERROR}
     * @throws NullPointerException if {@code statusTo} is {@code null}
     */
    Builder statusTo(StandardIo statusTo);

    /**
     * Selects which of this builder's output streams note messages should be
     * written to.
     *
     * @param noteTo which output stream note messages should be written to;
     *          must be {@code StandardIo.OUTPUT} or {@code StandardIo.ERROR}
     *
     * @return this builder
     *
     * @throws IllegalArgumentException if {@code noteTo} is not
     *           {@code StandardIo.OUTPUT} or {@code StandardIo.ERROR}
     * @throws NullPointerException if {@code noteTo} is {@code null}
     */
    Builder noteTo(StandardIo noteTo);

    /**
     * Selects which of this builder's output streams debug messages should be
     * written to.
     *
     * @param debugTo which output stream debug messages should be written to;
     *          must be {@code StandardIo.OUTPUT} or {@code StandardIo.ERROR}
     *
     * @return this builder
     *
     * @throws IllegalArgumentException if {@code debugTo} is not
     *           {@code StandardIo.OUTPUT} or {@code StandardIo.ERROR}
     * @throws NullPointerException if {@code debugTo} is {@code null}
     */
    Builder debugTo(StandardIo debugTo);

    /**
     * Sets the prefix for status messages.
     * <p>
     * All status messages will be prefixed with the specified prefix.
     *
     * @param statusPrefix the prefix for status messages
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code statusPrefix} is {@code null}
     */
    Builder statusPrefix(String statusPrefix);

    /**
     * Sets the prefix for note messages.
     * <p>
     * All note messages will be prefixed with the specified prefix.
     *
     * @param notePrefix the prefix for note messages
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code notePrefix} is {@code null}
     */
    Builder notePrefix(String notePrefix);

    /**
     * Sets the prefix for debug messages.
     * <p>
     * All debug messages will be prefixed with the specified prefix.
     *
     * @param debugPrefix the prefix for debug messages
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code debugPrefix} is {@code null}
     */
    Builder debugPrefix(String debugPrefix);

    /**
     * Sets the prefix for warn messages.
     * <p>
     * All warn messages will be prefixed with the specified prefix.
     *
     * @param warnPrefix the prefix for warn messages
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code warnPrefix} is {@code null}
     */
    Builder warnPrefix(String warnPrefix);

    /**
     * Sets the prefix for error messages.
     * <p>
     * All error messages will be prefixed with the specified prefix.
     *
     * @param errorPrefix the prefix for error messages
     *
     * @return this builder
     *
     * @throws NullPointerException if {@code errorPrefix} is {@code null}
     */
    Builder errorPrefix(String errorPrefix);

    /**
     * Creates a new UI from this builder.
     *
     * @return a new UI
     */
    ConfigurableSensibleUi build();
  }
}
