/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.BufferedWriter;
import java.io.FileDescriptor;
import java.io.FileOutputStream;
import java.io.Flushable;
import java.io.IOError;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.nio.channels.Channels;
import java.nio.charset.Charset;
import java.util.List;
import java.util.ListIterator;
import java.util.Locale;
import java.util.Objects;

import org.imca_cat.prgm.internal.AppendableWriter;
import org.imca_cat.prgm.internal.ProgramOutputFormatter;
import org.imca_cat.prgm.internal.PromptChoicesParser;
import org.imca_cat.prgm.internal.UnbufferedWritableByteChannelWriter;
import org.imca_cat.prgm.internal.ValueHolder;

/**
 * A configurable text user interface.
 * <p>
 * This is the default implementation of the {@link ConfigurableSensibleUi}
 * interface (and the {@link SensibleUi} interface).
 * <p>
 * Use a {@link Builder Builder} to construct an instance of this class, or use
 * the {@link DefaultConfigurableSensibleUi#create()} static factory method to
 * construct a default instance.
 * <p>
 * Writing to the standard output stream and standard error stream
 * {@link Appendable} instances may throw an exception due to an I/O error. If
 * this is not the desired behavior, a different {@link Appendable} can be
 * supplied to the builder that prevents such exceptions from being thrown
 * (e.g., a {@link PrintWriter}).
 * <p>
 * If an exception stack trace is printed, the stream on which it is printed
 * will be flushed after printing it. Also, the {@linkplain Locale#getDefault()
 * system locale} will be used when printing the stack trace, not the locale of
 * this UI.
 * <p>
 * This class is conditionally thread-safe. All methods may be called
 * concurrently. Output to the standard output stream and standard error stream
 * is guarded by a lock to ensure output on a terminal is not corrupted (e.g.,
 * two messages character-interleaved on the same line). The
 * {@link #prompt(String, String)} and
 * {@link #promptChoice(String, int, String)} methods may block waiting for
 * user input if this UI has been created with a {@link ProgramConsole}. The
 * methods may also block if a client calls one of these methods and blocks,
 * and then a second client calls one of these methods from another thread; the
 * second client may block waiting for access to the lock. The standard output
 * stream and standard error stream are not guarded by the same lock as the
 * {@code ProgramConsole} used by the {@link #prompt(String, String)} and
 * {@link #promptChoice(String, int, String)} methods.
 */
public final class DefaultConfigurableSensibleUi implements ConfigurableSensibleUi {
  private final Builder builder;
  private final String programName;
  private final ProgramConsole console;
  private final ProgramOutputFormatter stdout;
  private final ProgramOutputFormatter stderr;
  private final boolean autoFlushStdout;
  private final boolean autoFlushStderr;
  private final Locale locale;
  private final boolean quiet;
  private final boolean verbose;
  private final boolean debug;
  private final boolean printStackTrace;
  private final boolean useLocalizedExceptionMessage;
  private final StandardIo statusTo;
  private final StandardIo noteTo;
  private final StandardIo debugTo;
  private final String statusPrefix;
  private final String notePrefix;
  private final String debugPrefix;
  private final String warnPrefix;
  private final String errorPrefix;
  private final Object stdoutAndStderrLock;

  private DefaultConfigurableSensibleUi(Builder b) {
    builder = new Builder(b);
    programName = b.programName;
    console = b.console.hasValue() ? b.console.value() : ConsoleProgramConsole.systemInstance();
    locale = (b.locale != null) ? b.locale : Locale.getDefault();
    if (b.stdout != null) {
      stdout = new ProgramOutputFormatter(b.stdout, locale, StandardIo.OUTPUT);
    } else {
      /* Fully-buffered. */
      stdout = new ProgramOutputFormatter(
          new BufferedWriter(new OutputStreamWriter(new FileOutputStream(FileDescriptor.out))),
          locale, StandardIo.OUTPUT);
    }
    if (b.stderr != null) {
      stderr = new ProgramOutputFormatter(b.stderr, locale, StandardIo.ERROR);
    } else {
      /* Unbuffered. */
      stderr = new ProgramOutputFormatter(
          new UnbufferedWritableByteChannelWriter(
              Channels.newChannel(new FileOutputStream(FileDescriptor.err))),
          locale, StandardIo.ERROR);
    }
    autoFlushStdout = b.autoFlushStdout.hasValue() ? b.autoFlushStdout.value()
        : ((console == null) ? false : true);
    autoFlushStderr = b.autoFlushStderr;
    boolean tmpQuiet = b.quiet;
    boolean tmpVerbose = b.verbose;
    boolean tmpDebug = b.debug;
    if (tmpDebug) {
      tmpQuiet = false;
      tmpVerbose = true;
    } else if (tmpQuiet && tmpVerbose) {
      /* They cancel out. */
      tmpQuiet = false;
      tmpVerbose = false;
    }
    quiet = tmpQuiet;
    verbose = tmpVerbose;
    debug = tmpDebug;
    printStackTrace = b.printStackTrace;
    useLocalizedExceptionMessage = b.useLocalizedExceptionMessage;
    statusTo = b.statusTo;
    noteTo = b.noteTo;
    debugTo = b.debugTo;
    statusPrefix = b.statusPrefix;
    notePrefix = b.notePrefix;
    debugPrefix = b.debugPrefix;
    String defaultWarnAndErrorPrefix = (programName != null) ? programName + ": " : "";
    warnPrefix = (b.warnPrefix != null) ? b.warnPrefix : defaultWarnAndErrorPrefix;
    errorPrefix = (b.errorPrefix != null) ? b.errorPrefix : defaultWarnAndErrorPrefix;
    stdoutAndStderrLock = new Object();
  }

  /**
   * {@inheritDoc}
   * <p>
   * This method may be used for output that should always be written to the
   * standard output stream regardless of the level flags and without a level
   * prefix.
   *
   * @throws java.util.IllegalFormatException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while writing to the
   *           standard output stream
   */
  @Override
  public void write(String format, Object... args) {
    synchronized (stdoutAndStderrLock) {
      stdout.format(format, args);
      if (autoFlushStdout) stdout.flush();
    }
  }

  /**
   *
   * {@inheritDoc}
   * <p>
   * This method may be used for output that should always be written to the
   * standard error stream regardless of the level flags and without a level
   * prefix.
   *
   * @throws java.util.IllegalFormatException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while writing to the
   *           standard error stream
   */
  @Override
  public void writeStderr(String format, Object... args) {
    synchronized (stdoutAndStderrLock) {
      stderr.format(format, args);
      if (autoFlushStderr) stderr.flush();
    }
  }

  /**
   * Writes the message, prefixed with the status prefix, to the underlying
   * stream if the {@code quiet} level flag is not set.
   *
   * @throws java.util.IllegalFormatException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying stream
   */
  @Override
  public void status(String format, Object... args) {
    if (quiet) return;
    synchronized (stdoutAndStderrLock) {
      ProgramOutputFormatter f = statusTo.equals(StandardIo.ERROR) ? stderr : stdout;
      boolean autoFlush = statusTo.equals(StandardIo.ERROR) ? autoFlushStderr : autoFlushStdout;
      f.format("%s", statusPrefix);
      f.format(format, args);
      if (autoFlush) f.flush();
    }
  }

  /**
   * Writes the message, prefixed with the note prefix, to the underlying
   * stream if the {@code verbose} level flag is set.
   *
   * @throws java.util.IllegalFormatException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying stream
   */
  @Override
  public void note(String format, Object... args) {
    if (!verbose) return;
    synchronized (stdoutAndStderrLock) {
      ProgramOutputFormatter f = noteTo.equals(StandardIo.ERROR) ? stderr : stdout;
      boolean autoFlush = noteTo.equals(StandardIo.ERROR) ? autoFlushStderr : autoFlushStdout;
      f.format("%s", notePrefix);
      f.format(format, args);
      if (autoFlush) f.flush();
    }
  }

  /**
   * Writes the message, prefixed with the debug prefix, to the underlying
   * stream if the {@code debug} level flag is set.
   *
   * @throws java.util.IllegalFormatException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying stream
   */
  @Override
  public void debug(String format, Object... args) {
    if (!debug) return;
    synchronized (stdoutAndStderrLock) {
      ProgramOutputFormatter f = debugTo.equals(StandardIo.ERROR) ? stderr : stdout;
      boolean autoFlush = debugTo.equals(StandardIo.ERROR) ? autoFlushStderr : autoFlushStdout;
      f.format("%s", debugPrefix);
      f.format(format, args);
      if (autoFlush) f.flush();
    }
  }

  /**
   * Writes the message, prefixed with the warn prefix, to the standard error
   * stream.
   * <p>
   * A causing/associated exception my be supplied in addition to the message
   * format and arguments by including it as the last argument in this method's
   * varargs argument. This behavior is specified in the
   * {@link SensibleUi#warn(String format, Object... args)} method
   * documentation.
   * <p>
   * If printing stack traces is enabled, to write a stack trace without any
   * message, pass an empty string as the {@code format} argument and the
   * exception as the second argument.
   * <p>
   * If a {@link ProgramDieException} is given as a causing/associated
   * exception, it is printed differently than other exceptions: a stack trace
   * will not be printed for it, and its fully qualified class name will be
   * omitted; just its message will be printed.
   * <p>
   * If an exception stack trace is printed, the standard error stream will be
   * flushed after printing it. Also, the {@linkplain Locale#getDefault()
   * system locale} will be used when printing the stack trace, not the locale
   * specified for this UI.
   *
   * @throws java.util.IllegalFormatException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while writing to the
   *           standard error stream
   */
  @Override
  public void warn(String format, Object... args) {
    Throwable t = exceptionAtEnd(args);
    if (t != null) {
      errorOrWarnException(warnPrefix, String.format(locale, format, args), t);
      return;
    }
    synchronized (stdoutAndStderrLock) {
      stderr.format("%s", warnPrefix);
      stderr.format(format, args);
      if (autoFlushStderr) stderr.flush();
    }
  }

  /**
   * Writes the message, prefixed with the error prefix, to the standard error
   * stream.
   * <p>
   * A causing/associated exception my be supplied in addition to the message
   * format and arguments by including it as the last argument in this method's
   * varargs argument. This behavior is specified in the
   * {@link SensibleUi#error(String format, Object... args)} method
   * documentation.
   * <p>
   * If printing stack traces is enabled, to write a stack trace without any
   * message, pass an empty string as the {@code format} argument and the
   * exception as the second argument.
   * <p>
   * If a {@link ProgramDieException} is given as a causing/associated
   * exception, it is printed differently than other exceptions: a stack trace
   * will not be printed for it, and its fully qualified class name will be
   * omitted; just its message will be printed.
   * <p>
   * If an exception stack trace is printed, the standard error stream will be
   * flushed after printing it. Also, the {@linkplain Locale#getDefault()
   * system locale} will be used when printing the stack trace, not the locale
   * specified for this UI.
   *
   * @throws java.util.IllegalFormatException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while writing to the
   *           standard error stream
   */
  @Override
  public void error(String format, Object... args) {
    Throwable t = exceptionAtEnd(args);
    if (t != null) {
      errorOrWarnException(errorPrefix, String.format(locale, format, args), t);
      return;
    }
    synchronized (stdoutAndStderrLock) {
      stderr.format("%s", errorPrefix);
      stderr.format(format, args);
      if (autoFlushStderr) stderr.flush();
    }
  }

  /**
   * @throws NullPointerException if any argument is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           standard error stream
   */
  private void errorOrWarnException(String prefix, String message, Throwable t) {
    Objects.requireNonNull(prefix, "prefix is null");
    if (t instanceof ProgramDieException) {
      errorOrWarnException(prefix, message, (ProgramDieException)t);
      return;
    }
    synchronized (stdoutAndStderrLock) {
      stderr.format("%s", prefix);
      if (!message.isEmpty()) stderr.format("%s: ", message);
      if (printStackTrace) {
        stderr.flush();
        PrintWriter w = new PrintWriter(new AppendableWriter(stderr.out()));
        /*
         * Likely uses the system locale and localized exception messages;
         * there's no way to change this.
         */
        t.printStackTrace(w);
        flushAndCheckError(w, StandardIo.ERROR);
      } else {
        if (useLocalizedExceptionMessage) {
          stderr.format("%s", t.toString());
        } else {
          stderr.format("%s", t.getClass().getName());
          String tMessage = t.getMessage();
          if (tMessage != null && !tMessage.isEmpty()) stderr.format(": %s", tMessage);
        }
        stderr.format("%n");
      }
      if (autoFlushStderr) stderr.flush();
    }
  }

  /**
   * @throws NullPointerException if any argument is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           standard error stream
   */
  private void errorOrWarnException(String prefix, String message, ProgramDieException e) {
    Objects.requireNonNull(prefix, "prefix is null");
    String eMessage = useLocalizedExceptionMessage ? e.getLocalizedMessage() : e.getMessage();
    if (eMessage == null || eMessage.isEmpty()) {
      /*
       * Don't write anything for a ProgramDieException with no detail
       * message; this gives clients a way to die quietly.
       */
      return;
    }
    synchronized (stdoutAndStderrLock) {
      stderr.format("%s", prefix);
      if (!message.isEmpty()) stderr.format("%s: ", message);
      stderr.format("%s%n", eMessage);
      if (autoFlushStderr) stderr.flush();
    }
  }

  /**
   * Causes any buffered output on the standard output stream, standard error
   * stream, and console to be written out.
   *
   * @throws StandardIoException if an error occurs while flushing the standard
   *           output stream, standard error stream, or the console
   */
  @Override
  public void flush() {
    StandardIoException e = null;
    boolean interrupted = false;
    synchronized (stdoutAndStderrLock) {
      e = flush(stdout, StandardIo.OUTPUT, e);
      interrupted = Thread.interrupted() || interrupted;
      e = flush(stderr, StandardIo.ERROR, e);
      interrupted = Thread.interrupted() || interrupted;
    }
    e = flush(console, StandardIo.CONSOLE, e);
    if (interrupted) Thread.currentThread().interrupt();
    if (e == null) return;
    throw e;
  }

  /**
   * {@inheritDoc}
   * <p>
   * This implementation calls {@link #flush()} before prompting the user on
   * the console.
   *
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while flushing the standard
   *           output stream, standard error stream, or console, or while
   *           writing/reading to/from the console
   */
  @Override
  public String prompt(String message, String defaultResponse) {
    Objects.requireNonNull(message, "message is null");
    if (console == null) return defaultResponse;
    flush();
    String result;
    try {
      result = console.readLine("%s ", message);
    } catch (IOError e) {
      throw new StandardIoException(StandardIo.CONSOLE, e);
    }
    return (result != null && !result.isEmpty()) ? result : defaultResponse;
  }

  /**
   * {@inheritDoc}
   * <p>
   * This implementation calls {@link #flush()} before prompting the user on
   * the console.
   *
   * @throws PromptChoicesSyntaxException {@inheritDoc}
   * @throws IllegalArgumentException {@inheritDoc}
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while flushing the standard
   *           output stream, standard error stream, or console, or while
   *           writing/reading to/from the console
   */
  @Override
  public int promptChoice(String message, int defaultResponse, String choices) {
    Objects.requireNonNull(message, "message is null");
    if (defaultResponse < 0) {
      throw new IllegalArgumentException("defaultResponse (" + defaultResponse + ") < 0");
    }
    Objects.requireNonNull(choices, "choices is null");
    if (console == null) return defaultResponse;
    flush();
    String response;
    try {
      response = console.readLine("%s ", message);
    } catch (IOError e) {
      throw new StandardIoException(StandardIo.CONSOLE, e);
    }
    if (response == null || response.isEmpty()) return defaultResponse;
    response = response.toLowerCase(locale);
    List<List<String>> parsedChoices = PromptChoicesParser.parse(choices);
    for (ListIterator<List<String>> i = parsedChoices.listIterator(); i.hasNext();) {
      if (i.next().contains(response)) return i.previousIndex();
    }
    return -1;
  }

  /**
   * {@inheritDoc}
   * <p>
   * This implementation calls {@link #flush()} before prompting the user on
   * the console.
   *
   * @throws NullPointerException {@inheritDoc}
   * @throws StandardIoException if an error occurs while flushing the standard
   *           output stream, standard error stream, or console, or while
   *           writing/reading to/from the console
   */
  @Override
  public char[] promptPassword(String message, char[] defaultPassword) {
    Objects.requireNonNull(message, "message is null");
    if (console == null) return defaultPassword;
    flush();
    char[] result;
    try {
      result = console.readPassword("%s ", message);
    } catch (IOError e) {
      throw new StandardIoException(StandardIo.CONSOLE, e);
    }
    return (result != null && result.length != 0) ? result : defaultPassword;
  }

  @Override
  public Builder toBuilder() {
    return new Builder(builder);
  }

  /**
   * Creates a new instance with {@link Builder Builder} defaults.
   *
   * @return a new instance
   */
  public static DefaultConfigurableSensibleUi create() {
    return (new Builder()).build();
  }

  private static Throwable exceptionAtEnd(Object... args) {
    if (args == null) return null;
    if (args.length == 0) return null;
    Object lastArg = args[args.length - 1];
    return (lastArg instanceof Throwable) ? (Throwable)lastArg : null;
  }

  private static StandardIoException flush(Flushable f, StandardIo io,
      StandardIoException original) {
    if (f == null) return original;
    try {
      f.flush();
    } catch (ThreadDeath th) {
      throw th;
    } catch (Throwable th) {
      StandardIoException thr = new StandardIoException(io, th);
      if (original == null) return thr;
      original.addSuppressed(thr);
    }
    return original;
  }

  private static void flushAndCheckError(PrintWriter w, StandardIo io) {
    /* The checkError() method performs a flush. */
    if (!w.checkError()) return;
    throw new StandardIoException(io);
  }

  /**
   * A <i>builder</i> for creating {@link DefaultConfigurableSensibleUi}
   * instances.
   */
  public static final class Builder implements ConfigurableSensibleUi.Builder {
    private String programName;
    private ValueHolder<ProgramConsole> console = new ValueHolder<>();
    private Appendable stdout;
    private Appendable stderr;
    private ValueHolder<Boolean> autoFlushStdout = new ValueHolder<>();
    private boolean autoFlushStderr = false;
    private Locale locale;
    private boolean quiet = false;
    private boolean verbose = false;
    private boolean debug = false;
    private boolean printStackTrace = true;
    private boolean useLocalizedExceptionMessage = false;
    private StandardIo statusTo = StandardIo.OUTPUT;
    private StandardIo noteTo = StandardIo.OUTPUT;
    private StandardIo debugTo = StandardIo.OUTPUT;
    private String statusPrefix = "";
    private String notePrefix = "";
    private String debugPrefix = "";
    private String warnPrefix;
    private String errorPrefix;

    /**
     * Constructs an instance.
     */
    public Builder() {
      super();
    }

    /**
     * Constructs an instance with the same values as the specified builder.
     *
     * @param b the builder whose values are to be placed in this builder
     */
    public Builder(Builder b) {
      super();
      programName = b.programName;
      if (b.console.hasValue()) console.value(b.console.value());
      stdout = b.stdout;
      stderr = b.stderr;
      if (b.autoFlushStdout.hasValue()) autoFlushStdout.value(b.autoFlushStdout.value());
      autoFlushStderr = b.autoFlushStderr;
      locale = b.locale;
      quiet = b.quiet;
      verbose = b.verbose;
      debug = b.debug;
      printStackTrace = b.printStackTrace;
      useLocalizedExceptionMessage = b.useLocalizedExceptionMessage;
      statusTo = b.statusTo;
      noteTo = b.noteTo;
      debugTo = b.debugTo;
      statusPrefix = b.statusPrefix;
      notePrefix = b.notePrefix;
      debugPrefix = b.debugPrefix;
      warnPrefix = b.warnPrefix;
      errorPrefix = b.errorPrefix;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is {@code null}.
     */
    @Override
    public Builder programName(String programName) {
      if (programName != null && programName.isEmpty()) {
        throw new IllegalArgumentException("Empty programName");
      }
      this.programName = programName;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is a {@link ProgramConsole} wrapping the system console
     * returned by {@link System#console()}. If {@code System.console()}
     * returns {@code null}, then the default is {@code null}.
     */
    @Override
    public Builder console(ProgramConsole console) {
      this.console.value(console);
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is a buffered {@link Writer} on the system standard output
     * stream using the {@linkplain Charset#defaultCharset() system character
     * set}.
     */
    @Override
    public Builder stdout(Appendable stdout) {
      this.stdout = Objects.requireNonNull(stdout, "stdout is null");
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is an unbuffered {@link Writer} on the system standard error
     * stream using the {@linkplain Charset#defaultCharset() system character
     * set}.
     */
    @Override
    public Builder stderr(Appendable stderr) {
      this.stderr = Objects.requireNonNull(stderr, "stderr is null");
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is {@code true} if the console of the built UI is present;
     * {@code false} otherwise.
     */
    @Override
    public Builder autoFlushStdout(boolean autoFlush) {
      autoFlushStdout.value(autoFlush);
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is {@code false}.
     */
    @Override
    public Builder autoFlushStderr(boolean autoFlush) {
      autoFlushStderr = autoFlush;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is the {@linkplain Locale#getDefault() system locale}.
     */
    @Override
    public Builder locale(Locale locale) {
      this.locale = Objects.requireNonNull(locale, "locale is null");
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * If the {@code debug} flag is {@code true}, the {@code quiet} flag will
     * become {@code false} and the {@code verbose} flag will become
     * {@code true}. If both the {@code quiet} flag and the {@code verbose}
     * flag are {@code true}, they will cancel out and become {@code false}.
     * <p>
     * The default is {@code false}.
     */
    @Override
    public Builder quiet(boolean quiet) {
      this.quiet = quiet;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * If the {@code debug} flag is {@code true}, the {@code quiet} flag will
     * become {@code false} and the {@code verbose} flag will become
     * {@code true}. If both the {@code quiet} flag and the {@code verbose}
     * flag are {@code true}, they will cancel out and become {@code false}.
     * <p>
     * The default is {@code false}.
     */
    @Override
    public Builder verbose(boolean verbose) {
      this.verbose = verbose;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * If the {@code debug} flag is {@code true}, the {@code quiet} flag will
     * become {@code false} and the {@code verbose} flag will become
     * {@code true}. If both the {@code quiet} flag and the {@code verbose}
     * flag are {@code true}, they will cancel out and become {@code false}.
     * <p>
     * The default is {@code false}.
     */
    @Override
    public Builder debug(boolean debug) {
      this.debug = debug;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * If stack traces are printed, they will be printed using the
     * {@linkplain Locale#getDefault() system locale}, not this builder's
     * locale.
     * <p>
     * The default is {@code true}.
     */
    @Override
    public Builder printStackTrace(boolean printStackTrace) {
      this.printStackTrace = printStackTrace;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is {@code false}.
     */
    @Override
    public Builder useLocalizedExceptionMessage(boolean useLocalizedExceptionMessage) {
      this.useLocalizedExceptionMessage = useLocalizedExceptionMessage;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is {@link StandardIo#OUTPUT}.
     */
    @Override
    public Builder statusTo(StandardIo statusTo) {
      Objects.requireNonNull(statusTo, "statusTo is null");
      if (!statusTo.equals(StandardIo.OUTPUT) && !statusTo.equals(StandardIo.ERROR)) {
        throw new IllegalArgumentException(
            "statusTo (" + statusTo + ") not " + StandardIo.OUTPUT + " or " + StandardIo.ERROR);
      }
      this.statusTo = statusTo;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is {@link StandardIo#OUTPUT}.
     */
    @Override
    public Builder noteTo(StandardIo noteTo) {
      Objects.requireNonNull(noteTo, "noteTo is null");
      if (!noteTo.equals(StandardIo.OUTPUT) && !noteTo.equals(StandardIo.ERROR)) {
        throw new IllegalArgumentException(
            "noteTo (" + noteTo + ") not " + StandardIo.OUTPUT + " or " + StandardIo.ERROR);
      }
      this.noteTo = noteTo;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is {@link StandardIo#OUTPUT}.
     */
    @Override
    public Builder debugTo(StandardIo debugTo) {
      Objects.requireNonNull(debugTo, "debugTo is null");
      if (!debugTo.equals(StandardIo.OUTPUT) && !debugTo.equals(StandardIo.ERROR)) {
        throw new IllegalArgumentException(
            "debugTo (" + debugTo + ") not " + StandardIo.OUTPUT + " or " + StandardIo.ERROR);
      }
      this.debugTo = debugTo;
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is the empty string.
     */
    @Override
    public Builder statusPrefix(String statusPrefix) {
      this.statusPrefix = Objects.requireNonNull(statusPrefix, "statusPrefix is null");
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is the empty string.
     */
    @Override
    public Builder notePrefix(String notePrefix) {
      this.notePrefix = Objects.requireNonNull(notePrefix, "notePrefix is null");
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is the empty string.
     */
    @Override
    public Builder debugPrefix(String debugPrefix) {
      this.debugPrefix = Objects.requireNonNull(debugPrefix, "debugPrefix is null");
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is the {@linkplain #programName(String) program name}
     * followed by a colon character and a space. If the program name is not
     * set, the default is the empty string.
     */
    @Override
    public Builder warnPrefix(String warnPrefix) {
      this.warnPrefix = Objects.requireNonNull(warnPrefix, "warnPrefix is null");
      return this;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The default is the {@linkplain #programName(String) program name}
     * followed by a colon character and a space. If the program name is not
     * set, the default is the empty string.
     */
    @Override
    public Builder errorPrefix(String errorPrefix) {
      this.errorPrefix = Objects.requireNonNull(errorPrefix, "errorPrefix is null");
      return this;
    }

    @Override
    public DefaultConfigurableSensibleUi build() {
      return new DefaultConfigurableSensibleUi(this);
    }
  }
}
