/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.Flushable;
import java.io.PrintWriter;
import java.io.Reader;

/**
 * A computer program console.
 * <p>
 * Since {@link java.io.Console} is final and not an interface, it is
 * impossible to replace it with a <i>test double</i>. This interface is
 * identical to that of {@code Console} in Java 6 and, by virtue of being an
 * interface, is easy to replace with a <i>test double</i>.
 */
public interface ProgramConsole extends Flushable {
  /**
   * Returns the underlying writer of this console.
   * <p>
   * This behaves exactly like {@link java.io.Console#writer()}.
   *
   * @return the underlying writer
   */
  public PrintWriter writer();

  /**
   * Returns the underlying reader of this console.
   * <p>
   * This behaves exactly like {@link java.io.Console#reader()}.
   *
   * @return the underlying reader
   */
  public Reader reader();

  /**
   * Prints the message.
   * <p>
   * This behaves exactly like
   * {@link java.io.Console#format(String, Object...)}.
   *
   * @param fmt the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @return this console
   *
   * @throws java.util.IllegalFormatException if {@code fmt} or {@code args} is
   *           illegal according to the rules of {@link java.util.Formatter}
   */
  public ProgramConsole format(String fmt, Object... args);

  /**
   * Prints the message.
   * <p>
   * This behaves exactly like
   * {@link java.io.Console#printf(String, Object...)}.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @return this console
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   */
  public ProgramConsole printf(String format, Object... args);

  /**
   * Prints a prompt, and reads a line of input.
   * <p>
   * This behaves exactly like
   * {@link java.io.Console#readLine(String, Object...)}.
   *
   * @param fmt the prompt format string as defined by
   *          {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @return the line read, not including the line terminator, or {@code null}
   *         if the end of the stream is reached
   *
   * @throws java.util.IllegalFormatException if {@code fmt} or {@code args} is
   *           illegal according to the rules of {@link java.util.Formatter}
   * @throws java.io.IOError if an I/O error occurs
   */
  public String readLine(String fmt, Object... args);

  /**
   * Reads a line of input.
   * <p>
   * This behaves exactly like {@link java.io.Console#readLine()}.
   *
   * @return the line read, not including the line terminator, or {@code null}
   *         if the end of the stream is reached
   *
   * @throws java.io.IOError if an I/O error occurs
   */
  public String readLine();

  /**
   * Prints a prompt, and reads a password with echoing turned off.
   * <p>
   * This behaves exactly like
   * {@link java.io.Console#readPassword(String, Object...)}.
   *
   * @param fmt the prompt format string as defined by
   *          {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @return the password read as a {@code char} array, not including the line
   *         terminator, or {@code null} if the end of the stream is reached
   *
   * @throws java.util.IllegalFormatException if {@code fmt} or {@code args} is
   *           illegal according to the rules of {@link java.util.Formatter}
   * @throws java.io.IOError if an I/O error occurs
   */
  public char[] readPassword(String fmt, Object... args);

  /**
   * Reads a password with echoing turned off.
   * <p>
   * This behaves exactly like {@link java.io.Console#readPassword()}.
   *
   * @return the password read as a {@code char} array, not including the line
   *         terminator, or {@code null} if the end of the stream is reached
   *
   * @throws java.io.IOError if an I/O error occurs
   */
  public char[] readPassword();

  /**
   * Ensures that any buffered output is written out.
   * <p>
   * This behaves exactly like {@link java.io.Console#flush()}.
   */
  public void flush();
}
