/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

/**
 * Indicates a {@link Program} has terminated.
 * <p>
 * Typically the reason message, if specified, is printed on the standard error
 * stream, and the VM, if its sole purpose is to run the program, is caused to
 * exit with the specified exit status.
 */
public final class ProgramDieException extends RuntimeException {
  private static final long serialVersionUID = 1L;

  private final String localizedMessage;
  private final int exitStatus;

  /**
   * Constructs an instance with the specified exit status.
   *
   * @param exitStatus the exit status of the program
   */
  public ProgramDieException(int exitStatus) {
    this(null, exitStatus);
  }

  /**
   * Constructs an instance with the specified reason message and exit status.
   *
   * @param message the reason message; may be {@code null}
   * @param exitStatus the exit status of the program
   */
  public ProgramDieException(String message, int exitStatus) {
    this(message, null, exitStatus);
  }

  /**
   * Constructs an instance with the specified reason message, localized reason
   * message, and exit status.
   *
   * @param message the reason message; may be {@code null}
   * @param localizedMessage the localized reason message; may be {@code null}
   * @param exitStatus the exit status of the program
   */
  public ProgramDieException(String message, String localizedMessage, int exitStatus) {
    super(message);
    this.localizedMessage = localizedMessage;
    this.exitStatus = exitStatus;
  }

  /**
   * Returns the localized reason message as specified at instance creation
   * time, falling back to the non-localized reason message if there is no
   * localized reason message.
   *
   * @return the localized reason message if present (i.e., not {@code null});
   *         otherwise the non-localized reason message (which may be
   *         {@code null})
   */
  @Override
  public String getLocalizedMessage() {
    return (localizedMessage != null) ? localizedMessage : getMessage();
  }

  /**
   * Returns the exit status of the program.
   *
   * @return the exit status
   */
  public int exitStatus() {
    return exitStatus;
  }
}
