/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import org.imca_cat.prgm.internal.Utilities;

/**
 * A class providing static utility methods related to programs.
 */
public final class Programs {
  private Programs() {
    throw new AssertionError();
  }

  /**
   * Runs the specified program and exits the VM with the exit status of the
   * program.
   * <p>
   * The program runs in a separate thread, and that thread is interrupted if a
   * VM shutdown is initiated.
   * <p>
   * The exit status used to exit the VM is the exit status returned by the
   * program's {@link Program#run() run()} method. If that method throws an
   * exception, the exit status used is the exit status returned by the
   * program's {@link Program#exitStatus(Throwable) exitStatus(Throwable)}
   * method when passed the exception as its argument. If that method throws an
   * exception, {@link Program#EXIT_FAILURE} is used.
   * <p>
   * Before exiting the VM, the program's {@link Program#flush() flush()}
   * method is called to ensure any buffered output gets written to underlying
   * standard I/O destinations.
   * <p>
   * NOTE: This method should only be called when the sole purpose of the VM is
   * to run the specified program. This method does things like registering a
   * VM shutdown hook and calling {@link System#exit(int)} to exit the VM when
   * the program terminates. This method normally does not return.
   *
   * @param p the program to run
   */
  public static void runAndExitVm(Program p) {
    Thread t = null;
    int exitStatus = Program.EXIT_FAILURE;
    try {
      ProgramRunner r = new ProgramRunner(p);
      t = registerVmShutdownHook(r);
      exitStatus = r.run();
    } catch (ThreadDeath th) {
      throw th;
    } catch (Throwable th) {
      printError(p, th);
      exitStatus = exitStatus(p, th);
      Utilities.flushSilently(p);
    } finally {
      unregisterVmShutdownHookSilently(t);
    }
    System.exit(exitStatus);
  }

  /**
   * Enables the VM's headless mode (i.e., the mode that does not require a
   * display, keyboard, nor mouse).
   * <p>
   * This is a convenience method that is equivalent to:
   *
   * <pre>
   * System.setProperty("java.awt.headless", "true");
   * </pre>
   */
  public static void enableVmHeadlessMode() {
    System.setProperty("java.awt.headless", "true");
  }

  private static Thread registerVmShutdownHook(final ProgramRunner r) throws IllegalStateException {
    Thread result = new Thread("ProgramShutdownHook") {
      @Override
      public void run() {
        try {
          r.stop();
          for (;;) {
            try {
              r.await();
              break;
            } catch (InterruptedException e) {
              /* IGNORED */
            }
          }
        } catch (ThreadDeath th) {
          throw th;
        } catch (Throwable th) {
          /* IGNORED */
        }
      }
    };
    result.setPriority(Thread.NORM_PRIORITY);
    Runtime.getRuntime().addShutdownHook(result);
    return result;
  }

  private static void unregisterVmShutdownHookSilently(Thread t) {
    if (t == null) return;
    try {
      Runtime.getRuntime().removeShutdownHook(t);
    } catch (ThreadDeath th) {
      throw th;
    } catch (Throwable th) {
      /* IGNORED */
    }
  }

  private static void printError(Program p, Throwable t) {
    try {
      p.printError(t);
    } catch (ThreadDeath th) {
      throw th;
    } catch (Throwable th) {
      /* IGNORED */
    }
  }

  private static int exitStatus(Program p, Throwable t) {
    try {
      return p.exitStatus(t);
    } catch (ThreadDeath th) {
      throw th;
    } catch (Throwable th) {
      return Program.EXIT_FAILURE;
    }
  }

  /**
   * Runs (and flushes) a {@link Program} in its own thread.
   */
  private static class ProgramRunner {
    private final ExecutorService service;
    private final CountDownLatch start;
    private final Future<Integer> future;

    public ProgramRunner(Program p) {
      super();
      service = Executors.newCachedThreadPool();
      start = new CountDownLatch(1);
      future = service.submit(newCallable(p, start));
    }

    public int run() throws Throwable {
      start.countDown();
      int result = Program.EXIT_FAILURE;
      Throwable t = null;
      try {
        result = future.get();
      } catch (CancellationException e) {
        t = new InterruptedException();
        t.initCause(e);
      } catch (ExecutionException e) {
        Throwable cause = e.getCause();
        /*
         * If the cause is a ThreadDeath, wrap it in an Error so that it
         * is not caught by the client and considered to indicate that the
         * client's thread has been stopped by a call to Thread.stop().
         */
        t = (cause instanceof ThreadDeath) ? new Error(cause) : cause;
      }
      stop();
      try {
        await();
      } catch (InterruptedException e) {
        if (t == null) {
          t = e;
        } else {
          t.addSuppressed(e);
          Thread.currentThread().interrupt();
        }
      }
      if (t == null) return result;
      throw t;
    }

    public void stop() {
      future.cancel(true);
      service.shutdown();
    }

    public void await() throws InterruptedException {
      service.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
    }

    private static Callable<Integer> newCallable(final Program p, final CountDownLatch start) {
      return new Callable<Integer>() {
        @Override
        public Integer call() throws Exception {
          start.await();
          int exitStatus = p.run();
          p.flush();
          return exitStatus;
        }
      };
    }
  }
}
