/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

/**
 * Indicates a syntax error in a UI prompt-choices string.
 *
 * @see SensibleUi#promptChoice(String, int, String)
 */
public final class PromptChoicesSyntaxException extends IllegalArgumentException {
  private static final long serialVersionUID = 1L;

  private final int errorOffset;

  /**
   * Constructs an instance with the specified error offset.
   *
   * @param errorOffset the zero-based character offset from the beginning of
   *          the choices string to the location of the error
   */
  public PromptChoicesSyntaxException(int errorOffset) {
    this("Error offset: " + errorOffset, errorOffset);
  }

  /**
   * Constructs an instance with the specified detail message and error offset.
   *
   * @param message the {@linkplain #getMessage() detail message} of this
   *          exception; may be {@code null}
   * @param errorOffset the zero-based character offset from the beginning of
   *          the choices string to the location of the error
   */
  public PromptChoicesSyntaxException(String message, int errorOffset) {
    super(message);
    this.errorOffset = errorOffset;
  }

  /**
   * Constructs an instance with the specified error offset and cause.
   *
   * @param errorOffset the zero-based character offset from the beginning of
   *          the choices string to the location of the error
   * @param cause the {@linkplain Throwable#getCause() cause} of this
   *          exception; may be {@code null}
   */
  public PromptChoicesSyntaxException(int errorOffset, Throwable cause) {
    this("Error offset: " + errorOffset, errorOffset, cause);
  }

  /**
   * Constructs an instance with the specified detail message, error offset,
   * and cause.
   *
   * @param message the {@linkplain #getMessage() detail message} of this
   *          exception; may be {@code null}
   * @param errorOffset the zero-based character offset from the beginning of
   *          the choices string to the location of the error
   * @param cause the {@linkplain Throwable#getCause() cause} of this
   *          exception; may be {@code null}
   */
  public PromptChoicesSyntaxException(String message, int errorOffset, Throwable cause) {
    super(message, cause);
    this.errorOffset = errorOffset;
  }

  /**
   * Returns the zero-based character position of the syntax error.
   *
   * @return the error offset
   */
  public int errorOffset() {
    return errorOffset;
  }
}
