/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.Flushable;

/**
 * A text user interface for a {@link SensibleProgram}.
 * <p>
 * This interface contains methods for writing output to, and prompting for
 * input from, the user. The methods for writing output allow for writing
 * output at different logical levels, similar in concept to levels in a
 * logging framework. A program using this UI would typically use the
 * {@code write*} methods for output that should always be written, the
 * {@code status}, {@code note}, and {@code debug} methods for messages to be
 * written at the <i>status</i>, <i>note</i>, and <i>debug</i> levels,
 * respectively, in ascending verbosity order, and the {@code warn} and
 * {@code error} methods for messages to be written at the <i>warn</i> and
 * <i>error</i> levels, respectively.
 * <p>
 * This interface was inspired by that of the {@code ui} object from
 * <a href="https://www.mercurial-scm.org/wiki/MercurialApi">Mercurial's
 * internal API</a>.
 */
public interface SensibleUi extends Flushable {
  /**
   * Writes the message to the standard output stream.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   * @throws NullPointerException if {@code format} is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying standard output stream
   */
  void write(String format, Object... args);

  /**
   * Writes the message to the standard error stream.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   * @throws NullPointerException if {@code format} is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying standard error stream
   */
  void writeStderr(String format, Object... args);

  /**
   * Writes the message at the <i>status</i> level.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   * @throws NullPointerException if {@code format} is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying standard I/O destination
   */
  void status(String format, Object... args);

  /**
   * Writes the message at the <i>note</i> level.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   * @throws NullPointerException if {@code format} is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying standard I/O destination
   */
  void note(String format, Object... args);

  /**
   * Writes the message at the <i>debug</i> level.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   * @throws NullPointerException if {@code format} is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying standard I/O destination
   */
  void debug(String format, Object... args);

  /**
   * Writes the message at the <i>warn</i> level.
   * <p>
   * A causing/associated exception may be supplied in addition to the message
   * format and arguments by including it as the last argument in this method's
   * varargs argument. The exception must be an instance of {@link Throwable}.
   * What is done with the exception is implementation dependent, but an
   * implementation could, for example, print the message followed by a stack
   * trace of the exception. If an exception needs to be supplied as the last
   * argument to this method but is intended as the last argument in the format
   * string argument list (e.g., as the argument for a {@code %s} format
   * specifier), the exception must be supplied as a {@code String} (e.g., by
   * calling its {@code toString()} method) or any appropriate type that is not
   * a {@code Throwable}.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments optionally followed by a
   *          {@code Throwable} cause/association
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   * @throws NullPointerException if {@code format} is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying standard I/O destination
   */
  void warn(String format, Object... args);

  /**
   * Writes the message at the <i>error</i> level.
   * <p>
   * A causing/associated exception may be supplied in addition to the message
   * format and arguments by including it as the last argument in this method's
   * varargs argument. The exception must be an instance of {@link Throwable}.
   * What is done with the exception is implementation dependent, but an
   * implementation could, for example, print the message followed by a stack
   * trace of the exception. If an exception needs to be supplied as the last
   * argument to this method but is intended as the last argument in the format
   * string argument list (e.g., as the argument for a {@code %s} format
   * specifier), the exception must be supplied as a {@code String} (e.g., by
   * calling its {@code toString()} method) or any appropriate type that is not
   * a {@code Throwable}.
   *
   * @param format the format string as defined by {@link java.util.Formatter}
   * @param args the format string arguments optionally followed by a
   *          {@code Throwable} cause/association
   *
   * @throws java.util.IllegalFormatException if {@code format} or {@code args}
   *           is illegal according to the rules of {@link java.util.Formatter}
   * @throws NullPointerException if {@code format} is {@code null}
   * @throws StandardIoException if an error occurs while writing to the
   *           underlying standard I/O destination
   */
  void error(String format, Object... args);

  /**
   * Causes any buffered output on the standard output stream, standard error
   * stream, and any implementation-dependent output destination (e.g., a
   * {@link ProgramConsole}) to be written out.
   *
   * @throws StandardIoException if an error occurs while flushing the
   *           underlying standard I/O destinations
   */
  @Override
  void flush();

  /**
   * Prompts the user with the specified prompt message and default response.
   * <p>
   * The prompt message should not include a trailing space character or other
   * trailing whitespace; appropriate whitespace or other characters will be
   * added automatically.
   * <p>
   * The default response will be returned if the console is not present, the
   * user simply presses enter (i.e., an empty string response), or the end of
   * the stream is reached.
   *
   * @param message the prompt message
   * @param defaultResponse the default response
   *
   * @return the user's response, if it could be obtained, without the trailing
   *         line separator; otherwise, the default response
   *
   * @throws NullPointerException if {@code message} is {@code null}
   * @throws StandardIoException if an error occurs while writing/reading
   *           to/from an underlying standard I/O destination/source
   */
  String prompt(String message, String defaultResponse);

  /**
   * Prompts the user with the specified prompt message, default response, and
   * choices.
   * <p>
   * The prompt message should not include a trailing space character or other
   * trailing whitespace; appropriate whitespace or other characters will be
   * added automatically.
   * <p>
   * The default response will be returned if the console is not present, the
   * user simply presses enter (i.e., an empty string response), or the end of
   * the stream is reached.
   * <p>
   * The choices are encoded in the choices format which is a comma-separated
   * list of choices. A choice may be split into alternatives that are all
   * considered to be equivalent by separating them with a slash. All choices
   * must be lowercase. A backslash escapes the character that follows it thus
   * allowing the encoding of the backslash, comma, and slash characters as
   * choices if needed. The user's response will be converted to lowercase
   * before searching the choices for a match.
   * <p>
   * Choices Format Examples:
   * <ul>
   * <li>Two choices allowing a response of "{@code y}" or "{@code n}" in
   * arbitrary case:
   *
   * <pre>
   * y,n</pre>
   * <li>Two choices allowing a response of "{@code y}", "{@code yes}",
   * "{@code n}", or "{@code no}" in arbitrary case:
   *
   * <pre>
   * y/yes,n/no</pre>
   * <li>Three choices allowing a response of "{@code y}", "{@code yes}",
   * "{@code yes, please}", "{@code n}", "{@code no}", "{@code no, thank you}",
   * "{@code h}", or "{@code help}" in arbitrary case:
   *
   * <pre>
   * y/yes/yes\, please,n/no/no\, thank you,h/help</pre>
   * </ul>
   *
   * @param message the prompt message
   * @param defaultResponse the zero-based index of the default response in the
   *          choices list; must be greater than or equal to zero
   * @param choices the choices encoded in the choices format; must be all
   *          lowercase
   *
   * @return the zero-based index of the user's response in the choices list
   *         (alternatives are considered to be the same choice), or {@code -1}
   *         if the user's response does not match any choice; otherwise (i.e.,
   *         the user's response could not be obtained), the default response
   *
   * @throws PromptChoicesSyntaxException if {@code choices} is malformed
   * @throws IllegalArgumentException if {@code defaultResponse} is less than
   *           zero
   * @throws NullPointerException if {@code message} or {@code choices} is
   *           {@code null}
   * @throws StandardIoException if an error occurs while writing/reading
   *           to/from an underlying standard I/O destination/source
   */
  int promptChoice(String message, int defaultResponse, String choices);

  /**
   * Prompts the user for a password with echoing turned off and with the
   * specified prompt message and default password.
   * <p>
   * The prompt message should not include a trailing space character or other
   * trailing whitespace; appropriate whitespace or other characters will be
   * added automatically.
   * <p>
   * The default password will be returned if the console is not present, the
   * user simply presses enter (i.e., an empty string password), or the end of
   * the stream is reached.
   *
   * @param message the prompt message
   * @param defaultPassword the default password
   *
   * @return the user's password, if it could be obtained, without the trailing
   *         line separator; otherwise, the default password
   *
   * @throws NullPointerException if {@code message} is {@code null}
   * @throws StandardIoException if an error occurs while writing/reading
   *           to/from an underlying standard I/O destination/source
   */
  char[] promptPassword(String message, char[] defaultPassword);
}