/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm.internal;

import java.io.Closeable;
import java.io.Flushable;
import java.io.InterruptedIOException;
import java.nio.channels.ClosedByInterruptException;
import java.util.Formatter;
import java.util.FormatterClosedException;
import java.util.Locale;
import java.util.Objects;

import org.imca_cat.prgm.StandardIoException;
import org.imca_cat.prgm.StandardIo;

/**
 * A formatter like {@link Formatter} but that throws a
 * {@link StandardIoException} if an {@link IOException} occurs.
 * <p>
 * Instead of storing the {@code IOException} and making it available to the
 * client like {@code Formatter} does, a {@code StandardIoException} is thrown
 * when an {@code IOException} occurs. This also means that the thrown
 * exception does not persist with no way to clear it like is the case for a
 * {@code Formatter}; the exception is thrown when it occurs. If the underlying
 * {@link Appendable} recovers, then writing to it can proceed without this
 * formatter throwing an exception.
 * <p>
 * This class is not thread-safe.
 */
public final class ProgramOutputFormatter implements Closeable, Flushable, AutoCloseable {
  private Formatter formatter;
  private final StandardIo io;
  private boolean closed;

  public ProgramOutputFormatter(Appendable a, StandardIo io) {
    this(a, Locale.getDefault(), io);
  }

  public ProgramOutputFormatter(Appendable a, Locale l, StandardIo io) {
    super();
    formatter = new Formatter(Objects.requireNonNull(a, "a is null"),
        Objects.requireNonNull(l, "l is null"));
    this.io = Objects.requireNonNull(io, "io is null");
    closed = false;
  }

  /**
   * @throws FormatterClosedException if this formatter is closed
   * @throws StandardIoException if an I/O error occurs
   */
  public void flush() {
    if (closed) throw new FormatterClosedException();
    formatter.flush();
    checkIOException();
  }

  /**
   * @throws StandardIoException if an I/O error occurs
   */
  public void close() {
    if (closed) return;
    formatter.close();
    checkIOException();
    closed = true;
  }

  /**
   * @throws FormatterClosedException if this formatter is closed
   * @throws StandardIoException if an I/O error occurs
   */
  public ProgramOutputFormatter format(String format, Object... args) {
    if (closed) throw new FormatterClosedException();
    formatter.format(format, args);
    checkIOException();
    return this;
  }

  /**
   * @throws FormatterClosedException if this formatter is closed
   * @throws StandardIoException if an I/O error occurs
   */
  public ProgramOutputFormatter format(Locale l, String format, Object... args) {
    if (closed) throw new FormatterClosedException();
    formatter.format(l, format, args);
    checkIOException();
    return this;
  }

  /**
   * @throws FormatterClosedException if this formatter is closed
   */
  public Locale locale() {
    if (closed) throw new FormatterClosedException();
    return formatter.locale();
  }

  /**
   * @throws FormatterClosedException if this formatter is closed
   */
  public Appendable out() {
    if (closed) throw new FormatterClosedException();
    return formatter.out();
  }

  private void checkIOException() {
    if (formatter.ioException() == null) return;
    if (formatter.ioException() instanceof ClosedByInterruptException ||
        formatter.ioException() instanceof InterruptedIOException) {
      resetFormatter();
      Thread.currentThread().interrupt();
      return;
    }
    StandardIoException e = new StandardIoException(io, formatter.ioException());
    resetFormatter();
    throw e;
  }

  private void resetFormatter() {
    formatter = new Formatter(formatter.out(), formatter.locale());
  }
}
