/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm.internal;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import org.imca_cat.prgm.PromptChoicesSyntaxException;

/**
 * A parser for an encoded list of prompt choices.
 * <p>
 * Choices are encoded in the choices format which is a comma-separated list of
 * choices. A choice may be split into alternatives that are all considered to
 * be equivalent by separating them with a slash. All choices must be
 * lowercase. A backslash escapes the character that follows it thus allowing
 * the encoding of the backslash, comma, and slash characters as choices if
 * needed.
 * <p>
 * Examples:
 * <ul>
 * <li>Two choices allowing a response of "{@code y}" or "{@code n}" in
 * arbitrary case:
 * <p>
 * <pre>
 * y,n</pre>
 * <li>Two choices allowing a response of "{@code y}", "{@code yes}",
 * "{@code n}", or "{@code no}" in arbitrary case:
 * <p>
 * <pre>
 * y/yes,n/no</pre>
 * <li>Three choices allowing a response of "{@code y}", "{@code yes}",
 * "{@code yes, please}", "{@code n}", "{@code no}", "{@code no, thank you}",
 * "{@code h}", or "{@code help}" in arbitrary case:
 * <p>
 * <pre>
 * y/yes/yes\, please,n/no/no\, thank you,h/help</pre>
 * </ul>
 */
public final class PromptChoicesParser {
  private static final int SLASH = '/';
  private static final int COMMA = ',';
  private static final int BACKSLASH = '\\';

  private final String in;

  /**
   * @throws NullPointerException if {@code in} is {@code null}
   */
  public PromptChoicesParser(String in) {
    this.in = Objects.requireNonNull(in);
  }

  /**
   * @throws PromptChoicesSyntaxException if there is a syntax error in the
   *           choices format input
   */
  public List<List<String>> parse() {
    if (in.isEmpty()) {
      throw new PromptChoicesSyntaxException("Empty", 0);
    }
    List<List<String>> result = new ArrayList<>();
    List<String> alternates = new ArrayList<>();
    StringBuilder alternate = new StringBuilder();
    boolean escaped = false;
    for (int cp, i = 0; i < in.length(); i += Character.charCount(cp)) {
      cp = in.codePointAt(i);
      if (escaped) {
        alternate.appendCodePoint(cp);
        escaped = false;
        continue;
      }
      switch (cp) {
        case SLASH:
          if (alternate.length() == 0) {
            throw new PromptChoicesSyntaxException("Empty alternative at " + i, i);
          }
          alternates.add(alternate.toString());
          alternate.delete(0, alternate.length());
          break;
        case COMMA:
          if (alternate.length() == 0) {
            throw new PromptChoicesSyntaxException("Empty alternative at " + i, i);
          }
          alternates.add(alternate.toString());
          alternate.delete(0, alternate.length());
          result.add(new ArrayList<String>(alternates));
          alternates.clear();
          break;
        case BACKSLASH:
          escaped = true;
          break;
        default:
          alternate.appendCodePoint(cp);
      }
    }
    if (escaped) {
      throw new PromptChoicesSyntaxException("Trailing escape", in.length() - 1);
    }
    if (alternate.length() > 0) {
      alternates.add(alternate.toString());
      result.add(new ArrayList<String>(alternates));
    }
    if (result.size() < 2) {
      throw new PromptChoicesSyntaxException("Number of choices < 2", in.length() - 1);
    }
    return result;
  }

  /**
   * @throws PromptChoicesSyntaxException if there is a syntax error in the
   *           choices format input
   */
  public static List<List<String>> parse(String in) {
    return (new PromptChoicesParser(in)).parse();
  }
}
